from contextlib import suppress
from typing import Callable

import pytz
from django.conf import settings
from django.db import connections
from django.db.models import Max, QuerySet
from django.template.loader import get_template
from jinjasql import JinjaSql  # type: ignore
from tqdm import tqdm
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.dataloader.utils import dictfetchall, dictfetchone
from wbcore.contrib.geography.models import Geography
from wbcore.utils.cache import mapping
from wbfdm.models.exchanges.exchanges import Exchange
from wbfdm.models.instruments.instruments import Instrument, InstrumentType

BATCH_SIZE: int = 10000
instrument_type_map = {
    "ADR": "american_depository_receipt",
    "CF": "close_ended_fund",
    "EQ": "equity",
    "ET": "exchange_traded_fund",
    "ETC": "exchange_traded_commodity",
    "ETN": "exchange_traded_note",
    "EWT": "warrant",
    "GDR": "global_depository_receipt",
    "GNSH": "genussschein",
    "INVT": "investment_trust",
    "NVDR": "non_voting_depository_receipt",
    "PREF": "preference_share",
    "UT": "abc",
}


def get_dataloader_parameters(
    quote_code: str | None = None,
    exchange_code: str | None = None,
    rkd_code: str | None = None,
    ibes_code: str | None = None,
) -> dict:
    dataloader_parameters = dict()
    if quote_code:
        dataloader_parameters["adjustments"] = {
            "path": "wbfdm.contrib.qa.dataloaders.adjustments.DatastreamAdjustmentsDataloader",
            "parameters": quote_code,
        }
        dataloader_parameters["corporate_actions"] = {
            "path": "wbfdm.contrib.qa.dataloaders.corporate_actions.DatastreamCorporateActionsDataloader",
            "parameters": quote_code,
        }
        if exchange_code:
            dataloader_parameters["market_data"] = {
                "path": "wbfdm.contrib.qa.dataloaders.market_data.DatastreamMarketDataDataloader",
                "parameters": [
                    quote_code,
                    exchange_code,
                ],
            }

    if rkd_code:
        dataloader_parameters["officers"] = {
            "path": "wbfdm.contrib.qa.dataloaders.officers.RKDOfficersDataloader",
            "parameters": rkd_code,
        }
        dataloader_parameters["statements"] = {
            "path": "wbfdm.contrib.qa.dataloaders.statements.RKDStatementsDataloader",
            "parameters": rkd_code,
        }

    if ibes_code:
        dataloader_parameters["financials"] = {
            "path": "wbfdm.contrib.qa.dataloaders.financials.IBESFinancialsDataloader",
            "parameters": ibes_code,
        }
        dataloader_parameters["reporting_dates"] = {
            "path": "wbfdm.contrib.qa.dataloaders.reporting_dates.IbesReportingDateDataloader",
            "parameters": ibes_code,
        }

    return dataloader_parameters


def convert_data(data, parent_source: str | None = None) -> dict:
    if len(data.keys()) == 0:
        return data
    data["country_id"] = mapping(Geography.countries, "code_2").get(data["country_id"])
    data["currency_id"] = mapping(Currency.objects).get(data["currency_id"])
    if instrument_type_id := data.pop("instrument_type_id"):
        try:
            data["instrument_type_id"] = mapping(InstrumentType.objects)[
                instrument_type_map.get(instrument_type_id, instrument_type_id)
            ]
        except KeyError:
            data["instrument_type_id"] = InstrumentType.objects.get_or_create(
                key=instrument_type_id.lower(),
                defaults={"name": instrument_type_id.title(), "short_name": instrument_type_id.title()},
            )[0].pk
    data["exchange_id"] = mapping(Exchange.objects, "source_id", source="qa-ds2-exchange").get(
        str(data["exchange_code"])
    )
    data["dl_parameters"] = get_dataloader_parameters(
        data.pop("quote_code"), data.pop("exchange_code"), data.pop("rkd_code"), data.pop("ibes_code")
    )
    data["additional_urls"] = urls.split(",") if (urls := data.get("additional_urls")) else []
    if parent_source:
        data["parent_id"] = mapping(Instrument.objects, "source_id", source=parent_source).get(str(data["parent_id"]))
    if (is_primary := data.get("is_primary")) and isinstance(is_primary, str):
        data["is_primary"] = is_primary.lower().strip() == "y"
    data.pop("phone")
    return data


def get_instrument_from_data(data, parent_source: str | None = None) -> Instrument | None:
    if len(data.keys()) == 0:
        return None
    instrument = Instrument(**convert_data(data, parent_source))
    instrument.pre_save()
    instrument.computed_str = instrument.compute_str()
    if (
        instrument.name and instrument.currency and instrument.instrument_type
    ):  # we return an instrument only if it contains the basic name, currency and type
        return instrument


def _bulk_create_instruments_chunk(instruments: list[Instrument], update_unique_identifiers: bool = False):
    update_fields = [
        "name",
        "dl_parameters",
        "description",
        "country",
        "currency",
        "parent",
        "primary_url",
        "additional_urls",
        "headquarter_address",
        "inception_date",
        "delisted_date",
        "exchange",
        "instrument_type",
        "computed_str",
        "search_vector",
        "trigram_search_vector",
        "is_primary",
    ]
    bulk_update_kwargs = {"update_conflicts": True}
    if update_unique_identifiers:
        update_fields.extend(
            [
                "refinitiv_identifier_code",
                "refinitiv_mnemonic_code",
                "isin",
                "sedol",
                "valoren",
                "cusip",
            ]
        )
        bulk_update_kwargs = {"ignore_conflicts": True}
    Instrument.objects.bulk_create(
        instruments, update_fields=update_fields, unique_fields=["source", "source_id"], **bulk_update_kwargs
    )


def update_instruments(sql_name: str, parent_source: str | None = None, context=None, debug: bool = False, **kwargs):
    template = get_template(sql_name, using="jinja").template  # type: ignore
    if context is None:
        context = {}
    query, params = JinjaSql(param_style="format").prepare_query(template, context)
    instruments = []

    gen = dictfetchall(connections["qa"].cursor().execute(query, params))
    if debug:
        gen = tqdm(gen)
    # we update in batch to be sure to not exhaust resources
    for row in gen:
        with suppress(TypeError):  # we don't fail if the given data doesn't satisfy the schema
            if instrument := get_instrument_from_data(row, parent_source=parent_source):
                instruments.append(instrument)

        if len(instruments) >= BATCH_SIZE:
            _bulk_create_instruments_chunk(instruments, **kwargs)
            instruments = []

    _bulk_create_instruments_chunk(instruments, **kwargs)


def update_or_create_item(
    external_id: int, get_item: Callable, source: str, parent_source: str | None = None
) -> Instrument:
    defaults = convert_data(get_item(external_id), parent_source=parent_source)

    dl_parameters = defaults.pop("dl_parameters", {})
    defaults.pop("source", None)
    defaults.pop("source_id", None)

    instrument, _ = Instrument.objects.update_or_create(
        source=source,
        source_id=external_id,
        defaults=defaults,
    )

    for key, value in dl_parameters.items():
        instrument.dl_parameters[key] = value

    instrument.save()

    return instrument


def get_item(external_id: int, template_name: str) -> dict:
    template = get_template(template_name, using="jinja").template  # type: ignore
    query, params = JinjaSql(param_style="format").prepare_query(template, {"source_id": external_id})
    return dictfetchone(connections["qa"].cursor().execute(query, params))


def trigger_partial_update(
    queryset: QuerySet,
    last_update_field: str,
    id_field: str,
    table_change_name: str,
    update: Callable,
    update_or_create_item: Callable,
):
    max_last_updated = queryset.aggregate(max_last_updated=Max(last_update_field)).get("max_last_updated")

    if max_last_updated is None:
        update()

    else:
        with connections["qa"].cursor() as cursor:
            cursor.execute(
                f"SELECT MAX(last_user_update) FROM sys.dm_db_index_usage_stats WHERE OBJECT_NAME(object_id) = '{table_change_name}'"
            )
            max_last_updated_qa = (
                pytz.timezone(settings.TIME_ZONE).localize(result[0]) if (result := cursor.fetchone()) else None
            )
            if max_last_updated_qa and max_last_updated_qa > max_last_updated:
                for _, security_id in cursor.execute(
                    f"SELECT UpdateFlag_, {id_field} FROM {table_change_name}"
                ).fetchall():
                    update_or_create_item(security_id)



import operator
from contextlib import suppress
from datetime import datetime
from functools import reduce
from typing import Any, Dict, Optional

from django.contrib.postgres.search import TrigramSimilarity
from django.core.exceptions import MultipleObjectsReturned
from django.db import IntegrityError, models
from django.db.models import Q
from wbcore.contrib.currency.import_export.handlers import CurrencyImportHandler
from wbcore.contrib.geography.models import Geography
from wbcore.contrib.io.exceptions import DeserializationError
from wbcore.contrib.io.imports import ImportExportHandler, ImportState

from wbfdm.models.exchanges import Exchange


class InstrumentLookup:
    ORDERED_KEYS = [
        "instrument_type",
        "isin",
        "refinitiv_identifier_code",
        "refinitiv_mnemonic_code",
        "identifier",
        "ticker",
    ]

    def __init__(self, model, trigram_similarity_min_score: float = 0.8):
        self.cache = {}
        self.model = model
        self.trigram_similarity_min_score = trigram_similarity_min_score

    @classmethod
    def _get_cache_key(cls, **data):
        if data:
            return "-".join([f"{k}:{data.get(k, None)}" for k in cls.ORDERED_KEYS if data.get(k, None) is not None])

    def _lookup_instrument(
        self,
        instrument_type=None,
        currency=None,
        exchange=None,
        name=None,
        only_investable_universe: bool = False,
        exact_lookup: bool = False,
        **identifiers,
    ):
        identifiers = {k: v for k, v in identifiers.items() if v is not None}
        # General lookup, we try to gracefully find the instrument based on all available identifier fields
        cache_key = self._get_cache_key(**identifiers)
        if cache_key and cache_key in self.cache:
            return self.cache[cache_key]

        instrument = None

        # We need to lookup ticker because some provider gives us ticker with or without space in it
        if only_investable_universe:
            instruments = self.model.objects.filter(is_investable_universe=True)
        else:
            instruments = self.model.objects.filter(is_security=True)

        # Try exact lookup on the filtered out universe
        for identifier_key in [
            "isin",
            "refinitiv_identifier_code",
            "refinitiv_mnemonic_code",
            "sedol",
            "cusip",
            "identifier",
        ]:
            if identifier := identifiers.get(identifier_key, None):
                with suppress(self.model.DoesNotExist, MultipleObjectsReturned):
                    identifier = str(identifier)
                    if (
                        identifier_key != "refinitiv_identifier_code"
                    ):  # RIC cannot be uppercased because its symbology implies meaning for lowercase characters
                        identifier = identifier.upper()
                    instrument = instruments.get(**{identifier_key: identifier})
                    break
        if not instrument and not exact_lookup:
            if instrument_type:
                if isinstance(instrument_type, str):  # in case we receive a key as instrument type
                    instruments = instruments.filter(instrument_type__key=instrument_type)
                else:  # in case we receive a primary key as instrument type
                    instruments = instruments.filter(instrument_type=instrument_type)
            lookup_fields = [
                "isin",
                "refinitiv_identifier_code",
                "refinitiv_mnemonic_code",
                "refinitiv_ticker",
                "identifier",
                "ticker",
            ]

            conditions = []
            for field in lookup_fields:
                if field_value := identifiers.get(field, None):
                    conditions.append(Q(**{f"{field}": field_value}))
                    if field == "isin":
                        conditions.append(Q(old_isins__contains=[field_value]))
            if conditions:
                instruments = instruments.filter(reduce(operator.or_, conditions))

            if currency:  # if currency is provides, we use it as validator
                instruments = instruments.filter(currency=currency)

            if exchange:
                instruments_tmp = instruments.filter(exchange=exchange)
                if instruments_tmp.exists():
                    instruments = instruments_tmp

            # last chance
            if name and instruments.count() > 1:
                instruments = instruments.annotate(similarity_score=TrigramSimilarity("name", name))
                if instruments.filter(similarity_score__gt=self.trigram_similarity_min_score).count() == 1:
                    instruments = instruments.filter(similarity_score__gt=self.trigram_similarity_min_score)
            if instruments.count() == 1:
                instrument = instruments.first()
            elif instrument_type and identifiers:
                # if instrument type was provided but we still didn't find the security, we try without the instrument type in case it was mislabeled
                instrument = self._lookup_instrument(
                    only_investable_universe=only_investable_universe,
                    exact_lookup=exact_lookup,
                    currency=currency,
                    exchange=exchange,
                    **identifiers,
                )
            if not instrument and name and identifiers:
                # Sometime, identifier provided emptied the queryset of possible instruments. In a last chance approach, we try to only look for security with the given name
                instrument = self._lookup_instrument(
                    only_investable_universe=only_investable_universe,
                    exact_lookup=exact_lookup,
                    instrument_type=instrument_type,
                    currency=currency,
                    exchange=exchange,
                    name=name,
                )
        if instrument:
            self.cache[cache_key] = instrument
        return instrument

    def lookup(self, only_security: bool = False, exact_lookup: bool = False, **lookup_kwargs):
        # To speed up lookup process, we try to get the quote from the investable universe first
        security = self._lookup_instrument(only_investable_universe=True, exact_lookup=exact_lookup, **lookup_kwargs)
        if not security:
            security = self._lookup_instrument(
                only_investable_universe=False, exact_lookup=exact_lookup, **lookup_kwargs
            )
        if not only_security and security:
            quotes = security.children.all()
            if quotes.exists():
                # We try to find the quote for that security based on the given exchange (if provided). Otherwise, we default to the security primary exchange
                exchange = lookup_kwargs.get("exchange")
                if exchange and quotes.filter(exchange=exchange).exists():
                    return quotes.filter(exchange=exchange).first()
                else:
                    return quotes.filter(is_primary=True).first()
        return security


class InstrumentImportHandler(ImportExportHandler):
    MODEL_APP_LABEL: str = "wbfdm.Instrument"
    allow_update_save_failure = True
    exclude_update_fields = ["name", "isin", "country"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instrument_lookup = InstrumentLookup(self.model)
        self.currency_handler = CurrencyImportHandler(self.import_source)

    def _deserialize(self, data: Dict[str, Any]):
        from wbfdm.models import Classification, InstrumentType

        if isinstance(data, int):
            data = dict(id=data)
        if data.get("currency", None):
            data["currency"] = self.currency_handler.process_object(data["currency"], read_only=True)[0]
        if instrument_type := data.get("instrument_type", None):
            if isinstance(instrument_type, str):
                data["instrument_type"] = InstrumentType.objects.get_or_create(
                    key=instrument_type,
                    defaults={"name": instrument_type.title(), "short_name": instrument_type.title()},
                )[0]
            elif isinstance(instrument_type, int):
                data["instrument_type"] = InstrumentType.objects.get(id=instrument_type)
        if data.get("country", None):
            data["country"] = Geography.dict_to_model(data["country"])
        if data.get("headquarter_city", None):
            data["headquarter_city"] = Geography.dict_to_model(data["headquarter_city"], level=Geography.Level.CITY)
        if inception_date := data.get("inception_date", None):
            data["inception_date"] = datetime.strptime(inception_date, "%Y-%m-%d").date()
        if classifications := data.pop("classifications", None):
            data["classifications"] = [Classification.dict_to_model(c) for c in classifications]
        if (exchange_data := data.pop("exchange", None)) and isinstance(exchange_data, dict):
            sanitized_dict = {k: v for k, v in exchange_data.items() if v is not None}
            if sanitized_dict:
                data["exchange"] = Exchange.dict_to_model(sanitized_dict)

    def _get_instance(
        self,
        data: Dict[str, Any],
        history: Optional[models.QuerySet] = None,
        only_security: bool = False,
        **kwargs,
    ) -> models.Model:
        if isinstance(data, self.model):
            return data
        if instrument_id := data.pop("id", None):
            try:
                return self.model.objects.get(id=instrument_id)
            except self.model.DoesNotExist:
                raise DeserializationError("Instrument id does not match an existing instrument")
        else:
            return self.instrument_lookup.lookup(only_security=only_security, **data)

    def _create_instance(self, data: Dict[str, Any], **kwargs) -> models.Model:
        from wbfdm.models.instruments.instruments import Instrument

        if not data.get("name", None):
            raise DeserializationError("Can't create an instrument without at least a name")
        classifications = data.pop("classifications", None)
        try:
            obj = Instrument.objects.create(
                **data,
                import_source=self.import_source,
            )
            if classifications:
                obj.classifications.set([c for c in classifications if c])
            return obj
        except IntegrityError:
            self.import_source.log += f"\nError while creating new instrument with data {data}"

        return None

    def process_object(self, underlying_instrument_data, **kwargs):
        if underlying_instrument_data:
            if isinstance(underlying_instrument_data, self.model):
                return underlying_instrument_data, ImportState.UNMODIFIED
            if isinstance(underlying_instrument_data, int):
                underlying_instrument_data = dict(id=underlying_instrument_data)
        return super().process_object(
            underlying_instrument_data,
            include_update_fields=[
                "isin",
                "ticker",
                "refinitiv_mnemonic_code",
                "refinitiv_identifier_code",
                "country",
                "exchange",
            ],
            **kwargs,
        )



from typing import TYPE_CHECKING

from pypika import Column, MSSQLQuery, Table
from pypika.terms import ValueWrapper

if TYPE_CHECKING:
    from pypika.queries import CreateQueryBuilder
    from pypika.terms import Term


def compile_source(source: str) -> "Term":
    return ValueWrapper(source).as_("source")


SOURCE_DS2 = compile_source("qa-ds2")
SOURCE_RKD = compile_source("qa-rkd")
SOURCE_IBES = compile_source("qa-ibes")


def create_table(tablename: "str", *columns: "Column") -> tuple[Table, "CreateQueryBuilder"]:
    table = Table(tablename)
    return table, MSSQLQuery.create_table(table).columns(*columns)
